// Generated by gir (https://github.com/gtk-rs/gir @ 88f37d4165a8)
// from ../builddir/rsvg (@ 5bf47e852751)
// from gir-files (@ 6cef10cd64df)
// DO NOT EDIT

#include "manual.h"
#include <stdalign.h>
#include <stdio.h>

int main() {
    printf("%s;%zu;%zu\n", "RsvgDimensionData", sizeof(RsvgDimensionData), alignof(RsvgDimensionData));
    printf("%s;%zu;%zu\n", "RsvgError", sizeof(RsvgError), alignof(RsvgError));
    printf("%s;%zu;%zu\n", "RsvgHandle", sizeof(RsvgHandle), alignof(RsvgHandle));
    printf("%s;%zu;%zu\n", "RsvgHandleClass", sizeof(RsvgHandleClass), alignof(RsvgHandleClass));
    printf("%s;%zu;%zu\n", "RsvgHandleFlags", sizeof(RsvgHandleFlags), alignof(RsvgHandleFlags));
    printf("%s;%zu;%zu\n", "RsvgLength", sizeof(RsvgLength), alignof(RsvgLength));
    printf("%s;%zu;%zu\n", "RsvgPositionData", sizeof(RsvgPositionData), alignof(RsvgPositionData));
    printf("%s;%zu;%zu\n", "RsvgRectangle", sizeof(RsvgRectangle), alignof(RsvgRectangle));
    printf("%s;%zu;%zu\n", "RsvgUnit", sizeof(RsvgUnit), alignof(RsvgUnit));
    return 0;
}
