// Generated by gir (https://github.com/gtk-rs/gir @ 88f37d4165a8)
// from ../builddir/rsvg (@ 5bf47e852751)
// from gir-files (@ 6cef10cd64df)
// DO NOT EDIT

#include "manual.h"
#include <stdio.h>

#define PRINT_CONSTANT(CONSTANT_NAME) \
    printf("%s;", #CONSTANT_NAME); \
    printf(_Generic((CONSTANT_NAME), \
                    char *: "%s", \
                    const char *: "%s", \
                    char: "%c", \
                    signed char: "%hhd", \
                    unsigned char: "%hhu", \
                    short int: "%hd", \
                    unsigned short int: "%hu", \
                    int: "%d", \
                    unsigned int: "%u", \
                    long: "%ld", \
                    unsigned long: "%lu", \
                    long long: "%lld", \
                    unsigned long long: "%llu", \
                    float: "%.6f", \
                    double: "%.6f", \
                    long double: "%ld"), \
           CONSTANT_NAME); \
    printf("\n");

int main() {
    PRINT_CONSTANT((gint) RSVG_ERROR_FAILED);
    PRINT_CONSTANT((guint) RSVG_HANDLE_FLAGS_NONE);
    PRINT_CONSTANT((guint) RSVG_HANDLE_FLAG_KEEP_IMAGE_DATA);
    PRINT_CONSTANT((guint) RSVG_HANDLE_FLAG_UNLIMITED);
    PRINT_CONSTANT((gint) RSVG_UNIT_CH);
    PRINT_CONSTANT((gint) RSVG_UNIT_CM);
    PRINT_CONSTANT((gint) RSVG_UNIT_EM);
    PRINT_CONSTANT((gint) RSVG_UNIT_EX);
    PRINT_CONSTANT((gint) RSVG_UNIT_IN);
    PRINT_CONSTANT((gint) RSVG_UNIT_MM);
    PRINT_CONSTANT((gint) RSVG_UNIT_PC);
    PRINT_CONSTANT((gint) RSVG_UNIT_PERCENT);
    PRINT_CONSTANT((gint) RSVG_UNIT_PT);
    PRINT_CONSTANT((gint) RSVG_UNIT_PX);
    return 0;
}
