// This file was generated by gir (https://github.com/gtk-rs/gir)
// from ../builddir/rsvg
// from gir-files
// DO NOT EDIT

mod handle;
pub use self::handle::Handle;

mod enums;
pub use self::enums::Error;
pub use self::enums::Unit;

mod flags;
pub use self::flags::HandleFlags;

pub(crate) mod traits {
    pub use super::handle::HandleExt;
}
