#[cfg(any(check, docsrs))]
fn probe_system_deps() {
    // do not probe libraries since librsvg-2 is generated by Meson
}

#[cfg(not(any(check, docsrs)))]
fn probe_system_deps() {
    // do not probe libraries since the docs.rs environment doesn't have them
    system_deps::Config::new().probe().unwrap();
}

fn main() {
    probe_system_deps();
}
